module Interval (Ivl, (+/-)) where

-- Ivl exported abstractly. We enforce the invariant that for
-- any interval "Ivl l u", l <= u by only allowing construction of
-- intervals that satisfy the invariant, and by ensuring that
-- all operations on intervals preserve the invariant. We also
-- enforce that intervals must be bounded.

-- See the solution notes for a discussion on a more general approach
-- allowing for unbounded intervals and handling of NaNs.

data Ivl = Ivl Double Double deriving (Show, Eq)

(+/-) :: Double -> Double -> Ivl
x +/- e
    | isNaN x || isNaN e           = error "Invalid interval (NaN)"
    | isInfinite x || isInfinite e = error "Unbounded interval"
    | e < 0                        = error "Negative error bound" 
    | otherwise                    = Ivl (x - e) (x + e)


instance Num Ivl where
    (Ivl lx ux) + (Ivl ly uy) = Ivl (lx + ly) (ux + uy)

    (Ivl lx ux) - (Ivl ly uy) = Ivl (lx - uy) (ux - ly)

    (Ivl lx ux) * (Ivl ly uy) = Ivl (minimum bs) (maximum bs)
        where
            bs = [lx * ly, lx * uy, ux * ly, ux * uy]

    abs ivl@(Ivl lx ux) | lx >= 0   = ivl
                        | ux < 0    = Ivl (-ux) (-lx)
                        | otherwise = Ivl 0 (max (abs lx) (abs ux))

    signum (Ivl l u) = Ivl (signum l) (signum u)

    -- A version that spells out the different cases. Equivalent to the above.
    -- Might be helpful to see what is going on.
    -- signum (Ivl lx ux) | lx > 0             = 1
    --                    | ux < 0             = -1
    --                    | lx == 0 && ux == 0 = 0
    --                    | lx == 0 && ux >  0 = Ivl 0 1
    --                    | lx < 0  && ux == 0 = Ivl (-1) 0
    --                    | otherwise          = Ivl (-1) 1

    fromInteger x = Ivl (fromInteger x) (fromInteger x)


-- We opt to disallow division by intervals including 0. See the solution
-- notes for a discussion on a more general approach.
  
instance Fractional Ivl where
    (Ivl lx ux) / (Ivl ly uy)
        | uy < 0 || ly > 0 = Ivl (minimum bs) (maximum bs)
        | otherwise        = error "Division by zero"
        where
            bs = [lx / ly, lx / uy, ux / ly, ux / uy]

    recip (Ivl lx ux)
        | ux < 0 || lx > 0 = Ivl (1/ux) (1/lx)
        | otherwise        = error "Division by zero"

    fromRational x = Ivl (fromRational x) (fromRational x)
