{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module MonoidFoldable2 where

import Data.Foldable
import Data.Semigroup
import Data.Monoid hiding ((<>))

type Position = (Double, Double)
newtype Length = Length {getLength :: Double} deriving (Show, Eq, Ord, Num, Fractional, Floating)
newtype Area   = Area {getArea :: Double} deriving (Show, Eq, Ord, Num, Fractional, Floating)


instance Bounded Length where
   minBound = 0
   maxBound = 1/0

instance Bounded Area where
   minBound = 0
   maxBound = 1/0

data Object
    = Rectangle {centre :: Position, width :: Length, height :: Length}
    | Circle    {centre :: Position, radius :: Length}


circumference :: Object -> Length
circumference (Rectangle {width = w, height = h}) = 2 * (w + h)
circumference (Circle {radius = r}) = 2 * pi * r

area :: Object -> Area
area (Rectangle {width = w, height = h}) = Area (getLength w * getLength h)
area (Circle {radius = r}) = Area (4 * pi * getLength r / 3)

data Drawing a
    = Element a
    | Group  [Drawing a]

data Statistics = 
    Statistics {
        avgArea          :: Area,
        avgCircumference :: Length,
        maxArea          :: Area,
        maxCircumference :: Length
    }
    deriving Show

data AccumStats =
    AccumStats {
        asCount            :: Sum Int,
        asSumArea          :: Sum Area,
        asSumCircumference :: Sum Length,
        asMaxArea          :: Max Area,
        asMaxCircumference :: Max Length
    }

instance Semigroup AccumStats where
    as1 <> as2 = AccumStats {
                     asCount            = asCount as1 <> asCount as2,
                     asSumArea          = asSumArea as1 <> asSumArea as2,
                     asSumCircumference = asSumCircumference as1
                                          <> asSumCircumference as2,
                     asMaxArea          = asMaxArea as1 <> asMaxArea as2,
                     asMaxCircumference = asMaxCircumference as1
                                          <> asMaxCircumference as2
                 }



instance Monoid AccumStats where
    mempty = AccumStats {
                 asCount            = mempty,
                 asSumArea          = mempty, 
                 asSumCircumference = mempty,
                 asMaxArea          = mempty,
                 asMaxCircumference = mempty
             }

    mappend = (<>)

instance Foldable Drawing where
    foldMap f (Element e) = f e
    foldMap f (Group ds)  = foldMap (foldMap f) ds

statistics :: Drawing Object -> Statistics
statistics d =
    Statistics {
       avgArea          = getSum (asSumArea as)
                          / fromIntegral (getSum (asCount as)),
       avgCircumference = getSum (asSumCircumference as)
                          / fromIntegral (getSum (asCount as)),
       maxArea          = getMax (asMaxArea as),
       maxCircumference = getMax (asMaxCircumference as)
    }
    where
        as = foldMap asObject d

asObject :: Object -> AccumStats
asObject o =
    AccumStats {
        asCount            = Sum 1,
        asSumArea          = Sum a, 
        asSumCircumference = Sum c,
        asMaxArea          = Max a,
        asMaxCircumference = Max c
    }
    where
        a = area o
        c = circumference o


testDrawing :: Drawing Object
testDrawing =
    Group [
        Group [
            Element (Rectangle {centre = (10,10), width = 5, height = 7}),
            Element (Circle {centre = (11,11), radius = 5})
        ],
        Element (Rectangle {centre = (5,5), width = 10, height = 5})
    ]
