-- Skew Binary Random-Access List

module SBRAL where

import Prelude hiding (head, tail, lookup, drop)

data Tree a = Leaf a | Node (Tree a) a (Tree a) deriving (Show,Eq)
type RList a = [(Int, Tree a)]


empty :: RList a
empty = []


isEmpty :: RList a -> Bool
isEmpty ts = null ts


cons :: a -> RList a -> RList a
cons x ((w1, t1) : (w2, t2) : wts) | w1 == w2 =
    (w1 * 2 + 1, Node t1 x t2) : wts
cons x wts = ((1, Leaf x) : wts)


head :: RList a -> a
head ((_, Leaf x)     : _) = x
head ((_, Node _ x _) : _) = x


tail :: RList a -> RList a
tail ((_, Leaf _): wts)        = wts
tail ((w, Node t1 _ t2) : wts) = (w', t1) : (w', t2) : wts
    where
        w' = w `div` 2


lookup :: Int -> RList a -> a
lookup i ((w, t) : wts) | i < w      = lookupTree i w t
                        | otherwise  = lookup (i - w) wts


lookupTree :: Int -> Int -> Tree a -> a
lookupTree _ _ (Leaf x) = x
lookupTree i w (Node t1 x t2)
    | i == 0    = x
    | i <= w'   = lookupTree (i - 1) w' t1
    | otherwise = lookupTree (i - w' - 1) w' t2
    where
        w' = w `div` 2

update :: Int -> a -> RList a -> RList a
update i x (wt@(w, t) : wts) | i < w      = (w, updateTree i x w t) : wts
                             | otherwise  = wt : update (i - w) x wts


updateTree :: Int -> a -> Int -> Tree a -> Tree a
updateTree _ x _ (Leaf _) = Leaf x
updateTree i x w (Node t1 y t2)
    | i == 0    = (Node t1 x t2)
    | i <= w'   = Node (updateTree (i - 1) x w' t1) y t2
    | otherwise = Node t1 y (updateTree (i - w' - 1) x w' t2)
    where
        w' = w `div` 2


-- Similar in style to given operations.
drop :: Int -> RList a -> RList a
drop 0 wts = wts
drop _ [] = []
drop n ((w, t) : wts)
    | n >= w    = drop (n - w) wts
    | otherwise = dropTree n w t wts


dropTree :: Int -> Int -> Tree a -> RList a -> RList a
dropTree _ _ (Leaf _) wts = wts
dropTree n w (Node t1 _ t2) wts
    | n == 1      = (w', t1) : (w', t2) : wts
    | n <= w' + 1 = dropTree (n - 1) w' t1 ((w', t2) : wts)
    | otherwise   = dropTree (n - 1 - w') w' t2 wts
    where
        w' = w `div` 2


-- More concise and perhaps easier to understand version.
drop2 :: Int -> RList a -> RList a
drop2 n wts | n <= 0 = wts
drop2 _ []           = []
drop2 n ((w, t) : wts)
    | n >= w    = drop2 (n - w) wts
    | otherwise =
        -- t cannot be a leaf as 0 < n < w 
        case t of
            Node t1 _ t2 -> drop2 (n - 1) ((w',t1) : (w',t2) : wts)
            where
                w' = w `div` 2

-- A variation on the above is to use (RList) tail: this avoids code
-- duplication and the need to establish that the first tree cannot be a leaf.
-- But of course, tail will carry out some redundant case analysis.
drop2' :: Int -> RList a -> RList a
drop2' n wts | n <= 0 = wts
drop2' _ []           = []
drop2' n wtwts@((w, t) : wts)
    | n >= w    = drop2' (n - w) wts
    | otherwise = drop2' (n - 1) (tail wtwts)


-- Version where the case analysis is manifestly total (and avoiding
-- "trivial" recursive calls when n = 0).
drop3 :: Int -> RList a -> RList a
drop3 n wts
    | n <= 0    = wts
    | otherwise = dropRec n wts
    where
        -- Invariant: n > 0
        dropRec _ [] = []
        dropRec n ((_, Leaf _) : wts)
            | n == 1    = wts
            | otherwise = dropRec (n - 1) wts
        dropRec n ((w, Node t1 _ t2) : wts)
            | n == w    = wts
            | n > w     = dropRec (n - w) wts
            | n == 1    = (w', t1) : (w', t2) : wts
            | otherwise = dropRec (n - 1) ((w', t1) : (w', t2) : wts)
            where
                w' = w `div` 2
